@@ -50,7 +50,11 @@ def annotate(self, idp):
     for constructor in self.symbol_decls[SYMBOL].constructors:
         constructor.symbol = (Symbol(name=constructor.name[1:])
                                 .annotate(self, {}))
-    self.symbol_decls[SYMBOL].translate()  # to populate .map
+
+    # populate .map of SYMBOL
+    for c in self.symbol_decls[SYMBOL].constructors:
+        assert not c.sorts
+        self.symbol_decls[SYMBOL].map[str(c)] = UnappliedSymbol.construct(c)
 Vocabulary.annotate = annotate
 
 
@@ -266,6 +270,7 @@ def rename_args(self, new_vars):
         arg, nv = self.definiendum.sub_exprs[i], list(new_vars.values())[i]
         if type(arg) == Variable \
         and arg.name in vars and arg.name not in new_vars:
+            vars.remove(arg.name)
             self.body = self.body.instantiate([arg], [nv])
             self.out = (self.out.instantiate([arg], [nv]) if self.out else
                         self.out)
@@ -276,6 +281,8 @@ def rename_args(self, new_vars):
             eq = AComparison.make('=', [nv, arg])
             self.body = AConjunction.make('∧', [eq, self.body])
 
+    self.check(not vars, f"Too many variables in head of rule: {self}")
+
     self.definiendum.sub_exprs = list(new_vars.values())
     self.quantees = [Quantee.make(v, v.sort) for v in new_vars.values()]
     return self
@@ -434,6 +441,8 @@ def annotate(self, idp):
         ('moveSymbols', Symbol(name=BOOL)),
         ('optionalPropagation', Symbol(name=BOOL)),
         ('manualPropagation', Symbol(name=BOOL)),
+        ('optionalRelevance', Symbol(name=BOOL)),
+        ('manualRelevance', Symbol(name=BOOL)),
         ('unit', open_types['unit']),
         ('heading', open_types['heading'])
     ]:
@@ -581,12 +590,12 @@ def annotate(self, voc, q_vars):
     self = AQuantification.annotate(self, voc, q_vars)
     self.type = self.sub_exprs[AAggregate.OUT].type if self.out else INT
 
-    assert not self.using_if, f"Internal error in Aggregate: {self}"
-    self.sub_exprs = [IfExpr.make(if_f=self.sub_exprs[AAggregate.CONDITION],
-            then_f=Number(number='1') if self.out is None else
-                    self.sub_exprs[AAggregate.OUT],
-            else_f=Number(number='0'))]
-    self.using_if = True
+    if not self.using_if:
+        self.sub_exprs = [IfExpr.make(if_f=self.sub_exprs[AAggregate.CONDITION],
+                then_f=Number(number='1') if self.out is None else
+                        self.sub_exprs[AAggregate.OUT],
+                else_f=Number(number='0'))]
+        self.using_if = True
     return self
 AAggregate.annotate = annotate
 AAggregate.annotate1 = AQuantification.annotate1
