@@ -147,7 +147,7 @@ class IDP(ASTNode):
         assert path.exists(file), f"Can't find {file}"
         with open(file, "r") as source:
             code = source.read()
-            cls.from_str(code)
+            return cls.from_str(code)
 
     @classmethod
     def from_str(cls, code:str) -> "IDP":
@@ -349,8 +349,6 @@ class TypeDeclaration(ASTNode):
 
         interpretation (SymbolInterpretation): the symbol interpretation
 
-        translated (Z3): the translation of the type in Z3
-
         map (Dict[string, Expression]): a mapping from code to Expression in range
     """
 
@@ -367,7 +365,6 @@ class TypeDeclaration(ASTNode):
         self.type = (self.name if type(enumeration) != Ranges else
                      enumeration.type)  # INT or REAL or DATE
 
-        self.translated = None
         self.range = None
         self.map = {}  # {String: constructor}
 
@@ -449,8 +446,6 @@ class SymbolDeclaration(ASTNode):
         self.unit: str = None
         self.heading: str = None
 
-        self.translated = None
-
         self.type = None  # a string
         self.domain = None  # all possible arguments
         self.range = None  # all possible values
@@ -877,9 +872,6 @@ class Tuple(ASTNode):
     def __repr__(self):
         return self.code
 
-    def translate(self):
-        return [arg.translate() for arg in self.args]
-
 class FunctionTuple(Tuple):
     def __init__(self, **kwargs):
         self.args = kwargs.pop('args')
@@ -898,19 +890,34 @@ class Ranges(Enumeration):
 
         tuples = []
         self.type = None
-        for x in self.elements:
-            if self.type == None:
-                self.type = x.fromI.type
-            if x.toI is None:
-                tuples.append(Tuple(args=[x.fromI]))
-            elif x.fromI.type == INT and x.toI.type == INT:
-                for i in range(x.fromI.py_value, x.toI.py_value + 1):
-                    tuples.append(Tuple(args=[Number(number=str(i))]))
-            elif x.fromI.type == DATE and x.toI.type == DATE:
-                for i in range(x.fromI.py_value, x.toI.py_value + 1):
-                    tuples.append(Tuple(args=[Number(number=str(i))])) #TODO1
-            else:
-                self.check(False, f"Can't have a range over reals")
+        if self.elements:
+            self.type = self.elements[0].fromI.type
+            for x in self.elements:
+                if x.fromI.type != self.type:
+                    if self.type in [INT, REAL] and x.fromI.type in [INT, REAL]:
+                        self.type = REAL  # convert to REAL
+                        tuples = [Tuple(args=[n.args[0].real()])
+                                  for n in tuples]
+                    else:
+                        self.check(False,
+                            f"incorrect value {x.fromI} for {self.type}")
+
+                if x.toI is None:
+                    tuples.append(Tuple(args=[x.fromI]))
+                elif self.type == INT and x.fromI.type == INT and x.toI.type == INT:
+                    for i in range(x.fromI.py_value, x.toI.py_value + 1):
+                        tuples.append(Tuple(args=[Number(number=str(i))]))
+                elif self.type == REAL and x.fromI.type == INT and x.toI.type == INT:
+                    for i in range(x.fromI.py_value, x.toI.py_value + 1):
+                        tuples.append(Tuple(args=[Number(number=str(float(i)))]))
+                elif self.type == REAL:
+                    self.check(False, f"Can't have a range over real: {x.fromI}..{x.toI}")
+                elif self.type == DATE and x.fromI.type == DATE and x.toI.type == DATE:
+                    for i in range(x.fromI.py_value, x.toI.py_value + 1):
+                        d = Date(iso=f"#{date.fromordinal(i).isoformat()}")
+                        tuples.append(Tuple(args=[d]))
+                else:
+                    self.check(False, f"Incorrect value {x.toI} for {self.type}")
         Enumeration.__init__(self, tuples=tuples)
 
     def contains(self, args, function, arity=None, rank=0, tuples=None):
@@ -953,6 +960,8 @@ class Display(ASTNode):
         self.moveSymbols = False
         self.optionalPropagation = False
         self.manualPropagation = False
+        self.optionalRelevance = False
+        self.manualRelevance = False
         self.name = "display"
 
     def run(self, idp):
@@ -1001,6 +1010,10 @@ class Display(ASTNode):
                     self.optionalPropagation = True
                 elif name == "manualPropagation":
                     self.manualPropagation = True
+                elif name == "optionalRelevance":
+                    self.optionalRelevance = True
+                elif name == "manualRelevance":
+                    self.manualRelevance = True
                 else:
                     raise IDPZ3Error(f"unknown display constraint:"
                                      f" {constraint}")
